/***********************************************************************
                          valparse.c  -  description
                             -------------------
    begin                : Wed Mar 21 2001
    copyright            : (C) 2001 by Georg P. Loczewski
    email                : gpl@lambda-bound.de
************************************************************************/

/************************************************************************
*                                                                       *
* This program is free software; you can redistribute it and/or modify  *
* it under the terms of the GNU General Public License as published by  *
* the Free Software Foundation; either version 2 of the License, or     *
* (at your option) any later version.                                   *
*                                                                       *
************************************************************************/
# include <stdio.h>
# include <string.h>
# include <ctype.h>

# include "arsc.h"


  static char buffer[1000];
  static char *current;		// current buffer pointer
  static FILE *ins;		// input stream

  static void print_primary_prompt ();
  static void print_secondary_prompt ();
  static void prime_buffer ();
  static int is_separator (int);
  static void skip_spaces ();
  static void skip_newlines ();

  static VALUE * read_value ();
  static VALUE * read_symbol ();
  static VALUE * read_string ();
  static VALUE * read_int ();
  static VALUE * read_list ();

void valparsinit (FILE * is)
{
  if (is == NULL) {
     ins = stdin;
  }
  else {
     ins = is;
  }
}

static void
print_primary_prompt ()
{
  printf ("\ninput-> ");
  fflush (stdout);
}

static void
print_secondary_prompt ()
{
  printf ("> ");
  fflush (stdout);
}


static void
prime_buffer ()
{
  char *p;

  p = fgets (buffer, 1000, ins);
  if (p == NULL) {		// eof return by fgets
    if (ins == stdin) {
      strcpy (buffer, "quit ");
    }
    else {
      strcpy (buffer, "*eof* ");
    }
    p = buffer;
  }

  // initialize the current pointer
  *(p + strlen (p) - 1) = '\0';	// erase the newline character
  current = buffer;
  skip_spaces ();
}

static int
is_separator (int c)
{
  switch (c) {
  case ' ':
  case '\t':
  case '\n':
  case EOF:
  case '\0':
  case '\'':
  case ';':
  case ')':
  case '(':
    return 1;
  }
  return 0;
}

static void
skip_spaces ()
{
  while ((*current == ' ') || (*current == '\t'))
    current++;
  if (*current == ';') {	// ignore comments
    while (*current)
      current++;
  }
}

static void
skip_newlines ()
{
  skip_spaces ();
  while (*current == '\0') {	// end of line
    print_secondary_prompt ();	// print prompt as long as buffer is
    prime_buffer ();		// empty
  }
}

VALUE *
read_with_prompt ()
{
  VALUE * obj;

  do {
    print_primary_prompt ();
    prime_buffer ();
  } while (*current == '\0');

  obj = read_value ();

  skip_spaces ();
  if (*current) {
    error ("unexpected characters at end of line:", current);
  }
  return obj;
}

VALUE *
read_without_prompt ()
{
  VALUE * obj;

  do {
    prime_buffer ();
  } while (*current == '\0');

  if ((strlen(current) == 5) &&
      (strcmp(current,"*eof*") == 0)) {
     obj = ars_alloc_value ();
     TYPE (obj) = EOFV;
  }
  else {
     obj = read_value ();
  }

  // make sure we are at and of line
  if (!EOFP(obj)) {
     skip_spaces ();
     if (*current) {
        error ("unexpected characters at end of line:", current);
     }
  }
  return obj;
}

static VALUE *
read_value ()
{
  VALUE *tval;

  if (*current == '(') {
    current++;
    return  read_list ();
  }
  else if (*current == '"') {
    current++;
    return read_string ();
  }
  else if (isdigit(*current)) {
    return read_int ();
  }
  else {
    tval = read_symbol ();
    prvalue(tval);
    printf("trace in read_val\n");
    return tval;
  }
}

static VALUE *
read_list ()
{
   VALUE *hd;
   VALUE *tl;

   skip_newlines ();

   if (*current == ')') {
     current++;
     return nil;
   }
   hd =  read_value ();
   tl =  read_list ();
   return cons(hd, tl);
}

static VALUE *
read_symbol ()
{
  char *symbuf, *p;
  int len;

  symbuf = (char *) GC_malloc(sizeof(char) * 41);
  for (p = symbuf,len = 0; 
       len < 40 && !is_separator (*current); 
       len++, *p++ = *current++);
  *p = '\0';
  return newSYMV(defname(symbuf));
}

static VALUE *
read_string ()
{
  char *strbuf, *p;
  int len;

  strbuf = (char *) GC_malloc(sizeof(char) * 81);
  for (p = strbuf,len = 0; 
       len < 40 && !(*current == '"'); 
       len++, *p++ = *current++);
  current++;
  *p = '\0';
  return newSTRV(strbuf, len);
}

static VALUE *
read_int ()
{
  char *ibuf, *p;
  int len;

  ibuf = (char *) GC_malloc(sizeof(char) * 21);
  for (p = ibuf,len = 0; 
       len < 20 && isdigit (*current); 
       len++, *p++ = *current++);
  *p = '\0';
  return newINTV(atoi(ibuf));
}

